import BaseController from '../common/base.controller'
import { router } from '../decorators/router'
import { CardGroup } from '../models/CardGroup'
import { ZError } from '../common/ZError'
import { Card } from '../models/subdoc/Card'
import { BaseConst } from '../constants/BaseConst'
import { BagItem, ItemType } from '../models/BagItem'
import ItemCtrl from '../logic/ItemCtrl'
import { error } from '../common/Debug'
import { ItemFuncCfg } from '../cfg/parsers/ItemFuncCfg'
import {
  getDefaultCardGroupByHero
} from '../dao/CardGroupDao'
import { ItemInfo } from '../logic/ItemDef'

export default class CardController extends BaseController {
  @router('post /api/:accountid/card_group/:heroid')
  async cardGroupList(req: any) {
    let { accountid, heroid } = req.params
    let query: any = { accountid: accountid, deleted: false, isdefault: false }
    if (heroid) {
      query.heroid = parseInt(heroid)
    }
    let records = await CardGroup.find(query)
    let result: any = []
    if (heroid) {
      result.push(getDefaultCardGroupByHero(heroid))
    } else {
      let user = req.user
      for(let [,hero] of user.heros) {
        result.push(getDefaultCardGroupByHero(hero.heroid))
      }
    }
    for (let record of records) {
      let data: any = record.toJson()
      result.push(data)
    }

    return result
  }

  @router('post /api/:accountid/group_info/:heroid/:gid')
  @router('get /svr/:accountid/group_info/:heroid/:gid')
  async cardGroupInfo(req: any) {
    let { accountid, heroid, gid } = req.params
    heroid = +heroid
    let record
    if (!gid || gid === 'default') {
      return getDefaultCardGroupByHero(heroid)
    }
    if (gid) {
      record = await CardGroup.findById(gid)
    }
    if (!record) {
      return getDefaultCardGroupByHero(heroid)
    }
    return record.toJson()
  }


  @router('post /api/:accountid/card_group/save/:gid')
  async saveCardGroup(req: any) {
    let { accountid, gid, heroid, selected, cards } = req.params
    if (!gid && !heroid) {
      throw new ZError(101, 'gid和heroid不能同时为空')
    }
    if (gid === 'default') {
      throw new ZError(106, '默认卡组不能修改')
    }
    if (selected && !heroid) {
      throw new ZError(102, '要将某卡组设为selected, 必须同时传heroid')
    }
    let record
    if (gid) {
      record = await CardGroup.findById(gid)
    }
    if (record && record.deleted) {
      throw new ZError(103, '卡组已被删除')
    }
    if (!record) {
      let count = await CardGroup.count({accountid, isdefault: false, deleted: false})
      let account = req.user
      if (count >= account.slot) {
        throw new ZError(105, '英雄可使用的卡槽数量不足')
      }
      record = new CardGroup()
      record.accountid = accountid
      record.isdefault = false
    }
    if (record.isdefault) {
      throw new ZError(104, '默认卡组不能修改')
    }
    if (selected) {
      await CardGroup.updateMany({
        accountid,
        heroid
      }, { $set: { 'selected': false } })
      record.selected = true
    }
    if (heroid) {
      record.heroid = heroid
    }
    if (cards) {
      let datas: Card[] = []
      for (let cardid of cards) {
        let card = new Card()
        card.cardid = cardid
        card.owned = true
        card.usetype = 0
        card.free = false
        datas.push(card)
      }
      record.cards = datas
    }
    await record.save()
    return record.toJson()
  }

  @router('post /api/:accountid/card_group/delete/:gid')
  async deleteCardGroup(req: any) {
    let { gid } = req.params
    if (gid === 'default') {
      throw new ZError(104, '默认卡组不能被删除')
    }
    let record = await CardGroup.findById(gid)
    if (!record) {
      throw new ZError(101, '未找到记录')
    }
    if (record && record.deleted) {
      throw new ZError(103, '卡组已被删除')
    }
    record.deleted = true
    record.deletetime = new Date()
    await record.save()
    return {}
  }

  @router('post /api/:accountid/card/draw')
  async drawCard(req: any) {
    let { count, accountid, itemid } = req.params
    count = count << 0
    itemid = itemid << 0
    if (count != 10 && count != 1) {
      throw new ZError(101, '参数不正确')
    }
    let cfg: ItemFuncCfg = count == 10 ?
      global.$cfg.get(BaseConst.ITEMFUNC).get(90002):
      global.$cfg.get(BaseConst.ITEMFUNC).get(90001)
    if (!cfg) {
      error(`抽卡 ${accountid} : [${count}, 无法找到抽卡配置`)
      throw new ZError(102, '无法找到对应的配置')
    }
    let itemInfo = ItemCtrl.getItemsByInfo(cfg.consume1)
    if (count == 1 && itemInfo.length == 0) {
      throw new ZError(103, '无法找到对应的配置')
    }
    if (count == 1 && itemInfo[0].id !== itemid) {
      throw new ZError(104, '解锁物品与配置不符')
    }
    if (count == 10) {
      if (itemid != itemInfo[0].id) {
        itemInfo = ItemCtrl.getItemsByInfo(cfg.consume2)
        if (itemInfo.length == 0 || itemInfo[0].id != itemid) {
          throw new ZError(104, '解锁物品与配置不符')
        }
      }
    }
    let record = await BagItem.findOne({accountid, itemid})
    if (!record || record.count < itemInfo[0].count) {
      throw new ZError(105, '解锁物品数量不足')
    }
    const items: ItemInfo[]  = []
    items.push(...ItemCtrl.getItemsByInfo(cfg.get))
    let account = req.user
    const results = await BagItem.parseCardItems(account, items, itemInfo)
    await BagItem.useItems(accountid, itemInfo, 'draw_card')
    await account.save()
    return results
  }

  @router('post /api/:accountid/card/destory')
  async destoryCards(req: any) {
    let {cards, accountid} = req.params
    let items: ItemInfo[] = []
    let itemUse: ItemInfo[] = []
    for (let card of cards) {
      let record = await BagItem.findOne({
        accountid,
        itemid: card.cardid,
        itemtype: ItemType.CARD
      })
      if (!record) {
        throw new ZError(11, 'item not found')
      }
      if (record.count < card.count) {
        throw new ZError(12, 'not enough item')
      }
      itemUse.push(new ItemInfo(card.cardid, card.count))
      let data = ItemCtrl.decomposeItemCard(card.cardid, card.count)
      items = items.concat(data)

    }
    let result: any = await BagItem.addItems(accountid, items, 'destory_card', itemUse)
    await BagItem.useItems(accountid, itemUse, 'destory_card')
    return items
  }
}
