import { GameRecord } from '../models/GameRecord'
import BaseController from '../common/base.controller'
import { role, router } from '../decorators/router'
import { RecordInfo, User } from '../models/User'
import { error } from '../common/Debug'
import { timeBeforeDay } from '../utils/time.util'
import { BaseConst } from '../constants/BaseConst'
import { MatchCfg } from '../cfg/parsers/MatchCfg'
import ItemCtrl from '../logic/ItemCtrl'
import { ItemInfo } from '../logic/ItemDef'
import { BagItem } from '../models/BagItem'
import { setGameEnd, updateRank } from '../service/rank'
import { ZError } from '../common/ZError'
import { ActRecord } from '../models/ActRecord'
import { parsePlayerMatchDate } from '../dao/RecordDao'

export default class RecordController extends BaseController {
  @role('anon')
  @router('post /api/:accountid/records')
  async recordList(req: any) {
    let { accountid } = req.params
    let records = await GameRecord.find({ 'players.accountid': accountid })
    let results: any = []
    for (let record of records) {
      results.push(record.toJSON())
    }
    return results
  }

  @router('post /api/:accountid/saverecord')
  @role('anon')
  @router('post /svr/record/save')
  async upload(req: any) {
    let record = new GameRecord(req.params)
    await record.save()
    if (record.mode == 2) {
      return {}
    }
    let matchid: number = record.mode == 1 ? (parseInt(record.matchid) || 100001) : 100002
    let cfg: MatchCfg = global.$cfg.get(BaseConst.MATCH).get(matchid)
    if (!cfg) {
      error(`match cfg not found: ${matchid}`)
    }

    let seasonData: any = {}
    let itemData: any = {}
    for (let player of record.players) {
      let user = await User.findById(player.accountid).exec()
      if (!user) {
        error(`save game record, account not found: ${ player.accountid }`)
        continue
      }
      await setGameEnd(player.accountid)
      await parsePlayerMatchDate(player, user, record.winner)
      await user.save()
      seasonData[player.playerid] = user.season_data
      /**
       * 获取胜利失败的物品
       */
      let items: ItemInfo[] = []
      let subType =  record.winner == player.team ? 'win' : 'loss'
      if (cfg && !player.escape) {
        if (record.winner == player.team && cfg.winget) {
          items = ItemCtrl.getItemsByInfo(cfg.winget);
        } else if (record.winner !== player.team && cfg.failget) {
          items = ItemCtrl.getItemsByInfo(cfg.failget);
        }
        await BagItem.addItems(player.accountid, items, 'game_' + subType, record.id)
      }
      itemData[player.playerid] = items
    }
    return {seasonData, itemData}
  }

  @router('post /api/:accountid/gamereward')
  async doubleReward(req: any) {
    let { accountid, roomid } = req.params
    let day = (new Date(0)).format('yyyy-MM-dd', true)
    let arecord = (await ActRecord.findOrCreate({
      accountid,
      actid: roomid,
      day
    })).doc
    if (arecord.count > 0) {
      throw new ZError(12, 'already goted')
    }
    arecord.count += 1
    let records = await GameRecord.find({'players.accountid': accountid, roomid})
    if (!records || records.length == 0) {
      throw new ZError(10, 'record not found')
    }
    let record = records[0]
    if (!record.matchid) {
      throw new ZError(11, 'current game mode has no double reward')
    }
    let isWin = false
    for (let pdata of record.players) {
      if (pdata.accountid == accountid) {
        if (pdata.team == record.winner) {
          isWin = true
        }
        break
      }
    }
    let cfg: MatchCfg = global.$cfg.get(BaseConst.MATCH).get(parseInt(record.matchid))
    if (!cfg) {
      error(`match cfg not found: ${record.matchid}`)
    }
    /**
     * 获取胜利失败的物品
     */
    let items: ItemInfo[] = []
    if (isWin) {
      items = ItemCtrl.getItemsByInfo(cfg.winget);
    } else {
      items = ItemCtrl.getItemsByInfo(cfg.failget);
    }
    await BagItem.addItems(accountid, items, 'double_reward', roomid)
    await arecord.save()
    return items
  }
}
