import {
  getModelForClass,
  modelOptions,
  mongoose,
  plugin,
  prop,
  Severity
} from '@typegoose/typegoose'
import { dbconn } from '../decorators/dbconn'
// @ts-ignore
import findOrCreate from 'mongoose-findorcreate'
import {
  Base,
  FindOrCreate,
  TimeStamps
} from '@typegoose/typegoose/lib/defaultClasses'
import { Hero } from './subdoc/Hero'
import { ZError } from '../common/ZError'
import { MoneyTypeConst } from '../constants/MoneyTypeConst'
import { timeBeforeDay } from '../utils/time.util'
import { BaseConst } from '../constants/BaseConst'
import { Card } from './subdoc/Card'

export class RecordInfo {
  @prop()
  public time: number
  /**
   * 游戏结果
   * @type {number} 1: 平局, 2: 赢 0:输
   */
  @prop()
  public status: number
}

interface AccountClass extends Base<string>, TimeStamps {
}

@dbconn()
// @index({ _id: 1}, { unique: true })
@plugin(findOrCreate)
@modelOptions({
  schemaOptions:
    { collection: 'account', timestamps: true },
  options: { allowMixed: Severity.ALLOW }
})
export class UserClass extends FindOrCreate {
  @prop()
  public _id: string
  @prop()
  public nickname: string
  @prop()
  public avatar: string

  @prop({ default: false })
  public locked: boolean
  @prop()
  public lockedtime?: number

  @prop()
  public comment?: string
  @prop()
  public lastLogin?: Date

  @prop({default: 0})
  public robot: number

  @prop({default: 1})
  public isnew: number
  /**
   * 已获得卡牌信息
   */
  @prop({ type: () => [Number] })
  public cards: number[]
  /**
   * 已获得卡牌信息, 20210121添加
   */
  @prop({ type: Card, default: new Map() })
  public cardMap: Map<string, Card>
  /**
   * 已解锁英雄信息
   */
  @prop({ type: Hero, default: new Map() })
  public heros: Map<string, Hero>
  /**
   * 货币信息, Deprecated
   * 具体的货币从BagItem中查找ItemType.MONEY的记录
   * coin: 金币
   * diamond: 钻石
   * hero_shard: 通用英雄碎片
   * hero_exp: 通用英雄经验
   * hero_shard_heroid: 英雄专用碎片
   * hero_exp_heroid: 英雄专用经验
   * card_scroll: 抽卡卷轴
   */
  @prop({ type: Number, default: new Map() })
  public moneys: Map<string, number>


  /**
   * 匹配: 胜利场数, 失败场数, 平局场数, 掉线场数
   */
  @prop({ type: Number, default: [0, 0, 0, 0] })
  public normal_stat: number[]
  /**
   * 赛季数据
   */
  @prop({ type: Number, default: new Map() })
  public season_data: Map<string, number>
  /**
   * 一个月内10场记录
   * @type {RecordInfo[]}
   */
  @prop({ _id: false, type: () => [RecordInfo] })
  public season_records: RecordInfo[]

  /**
   * 当前赛季排位分
   */
  @prop({ type: Number, default: -1 })
  public season_score: number
  /**
   * 所有未定义的信息, 供扩展
   */
  @prop({ type: mongoose.Schema.Types.Mixed })
  public extinfo: any

  public toJson() {
    return {
      accountid: this._id
    }
  }

  /**
   * 计算十场胜率
   */
  public getTenWinRate() {
    //twp=ten_wp*ten_wn-ten_lp*ten_ln-ten_ap*ten_an
    const fc = global.$cfg.get(BaseConst.FORMULA)
    let smallTime = timeBeforeDay(30)
    let records = this.season_records || []
    for (let r of records) {
      if (r.time < smallTime) {
        records.removeEx(r)
      }
    }
    let win = 0
    let lost = 0
    let total = records.length
    for (let record of records) {
      if (record.status == 2) {
        win++
      } else if (record.status == 0) {
        lost++
      }
    }
    return win * fc.get(70016).number / 100 - lost * fc.get(70017).number / 100 - (10 - total) * fc.get(70018).number / 100
  }
  /**
   * 获取匹配分
   * @return {number}
   */
  public getMatchScore() {
    const fc = global.$cfg.get(BaseConst.FORMULA)
    let twp = this.getTenWinRate()
    // mp=bm_mp+cp*(bm_twp+twp)
    return fc.get(70021).number + this.season_score * (fc.get(70020).number / 100 + twp)
  }

  public unlockHero(heroid: number, maxCount: number) {
    let hero = this.heros.get(heroid + '')
    let result = 0
    if (hero && !hero.trial && hero.count >= maxCount) {
      return result
    }
    if (!hero) {
      hero = new Hero()
      hero.heroid = heroid
      hero.free = false
      hero.trial = false
      hero.level = 1
      hero.trial_expire = 0
      hero.exp = 0
      hero.slot = 1
      hero.count = 1
      hero.time = Date.now()
    } else {
      hero.count += 1
    }
    result = hero.count
    this.heros.set(heroid + '', hero)
    return result
  }

  public async unlockCard(cardid: number, maxCount: number) {
    let card = this.cardMap.get(cardid + '')
    let result = 0
    if (card && card.count >= maxCount) {
      return result
    }
    if (!card) {
      card = new Card()
      card.cardid = cardid
      card.owned = true
      card.ban = false
      card.usetype = 0
      card.free = false
      card.count = 1
      card.free_expire = 0
      card.time = Date.now()
    } else {
      card.count += 1
    }
    this.cardMap.set(card.cardid + '', card)
    return card.count
  }

  public async tryHero(heroid: number) {
    if (this.heros.has(heroid + '')) {
      throw new ZError(102, '你已经解锁了该英雄')
    }
    let hero = new Hero()
    hero.heroid = heroid
    hero.free = false
    hero.trial = true
    hero.level = 1
    hero.count = 0
    hero.trial_expire = 0
    hero.exp = 0
    this.heros.set(heroid + '', hero)
    return hero
  }

  /**
   * 升级英雄
   * @param heroid
   * @param items
   */
  public async updateHero(heroid: number, items: string) {
    if (items.indexOf(':') < 0) {
      throw new ZError(105, 'items参数格式错误')
    }
    if (!this.heros.has(heroid + '')) {
      throw new ZError(109, '升级英雄不存在')
    }
    let hero = this.heros.get(heroid + '')
    if (hero.trial) {
      throw new ZError(110, '试玩英雄不可升级')
    }

    let moneys: [string, number][] = []
    if (items.indexOf('|')) {
      let arr = items.split('|')
      for (let str of arr) {
        if (!str) {
          continue
        }
        if (str.indexOf(':') < 0) {
          throw new ZError(105, 'items参数格式错误')
        }

        let subarr = str.split(':')
        if (!MoneyTypeConst.checkMoney(subarr[0])) {
          throw new ZError(106, `货币不存在: ${ subarr[0] }`)
        }
        moneys.push([subarr[0], parseInt(subarr[1])])
      }
    } else {
      let arr = items.split(':')
      if (!MoneyTypeConst.checkMoney(arr[0])) {
        throw new ZError(106, `货币不存在: ${ arr[0] }`)
      }
      moneys.push([arr[0], parseInt(arr[1])])
    }
    const expPreLvl = 100
    let totalExp = 0
    for (let data of moneys) {
      let money = data[0]
      if (MoneyTypeConst.isHeroMoney(money)) {
        let tmpHeroId = MoneyTypeConst.getHeroId(money)
        if (tmpHeroId != heroid) {
          throw new ZError(107, `指定英雄的货币与当前英雄id不匹配: ${ money }`)
        }
      }
      let count = this.moneys.get(money) || 0
      if (count < data[1]) {
        throw new ZError(108, `当前货币不够: ${ money }, in: ${ data[1] }, had: ${ count }`)
      }
      this.moneys.set(money, count - data[1])
      // TODO:: 根据配表计算代币->经验值的换算比例
      let exp = data[1]
      totalExp += exp
    }
    // TODO:: 根据配表读取英雄升级需要的经验值
    let lvlInc = (hero.exp + totalExp) / expPreLvl | 0
    let expRemain = (hero.exp + totalExp) % expPreLvl
    hero.level += lvlInc
    hero.exp = expRemain
    return {
      exp: hero.exp,
      level: hero.level,
      exp_gain: totalExp
    }
  }
}


export const User = getModelForClass(UserClass,
// @ts-ignore
  { existingConnection: UserClass['db'] })
