import { RedisClient } from '../redis/RedisClient'
import { BaseConst } from '../constants/BaseConst'

const MAX_TIME = 5000000000000

/**
 * 更新排行榜数据
 * @param {string} accountid
 * @param {number} score
 * @return {Promise<void>}
 */
export async function updateRank(accountid: string, score: number) {
  let scoreL = parseFloat(`${score | 0}.${MAX_TIME - Date.now()}`)
  await new RedisClient().zadd(BaseConst.RANK_SCORE, scoreL, accountid)
}

/**
 * 获取指定分数段的玩家
 * @param {number} min
 * @param {number} max
 * @return {Promise<unknown>}
 */
export async function usersByScore(min: number, max: number) {
  return await new RedisClient().zrangebyscore(BaseConst.RANK_SCORE, min, max)
}

/**
 * 设置在游戏中的状态
 * @param {string} accountid
 * @param seconds
 * @return {Promise<void>}
 */
export async function setGameing(accountid: string, seconds: number) {
  await new RedisClient().setex('gameing_' + accountid, (Date.now() / 1000 | 0) + '', seconds)
}

export async function setGameEnd(accountid: string) {
  await new RedisClient().expire('gameing_' + accountid, 0)
}
/**
 * 检查是否在游戏中
 * @param {string} accountid
 * @return {Promise<unknown>}
 */
export async function checkGameing(accountid: string) {
  return await new RedisClient().get('gameing_' +  accountid)
}

/**
 * 获取指定用户的排名
 * @param {string} accountid
 * @return {Promise<unknown>}
 */
export async function getAccountRank(accountid: string) {
  return await new RedisClient().zrevrank(BaseConst.RANK_SCORE, accountid)
}
/**
 * 获取指定用户的积分
 * @param {string} accountid
 * @return {Promise<unknown>}
 */
export async function getAccountScore(accountid: string) {
  return await new RedisClient().zscore(BaseConst.RANK_SCORE, accountid)
}

/**
 * 获取从start到end的玩家列表
 * @param {number} start
 * @param {number} end
 * @return {Promise<unknown>}
 */
export async function getRankList(start: number, end: number) {
  return await new RedisClient().zrevrange(BaseConst.RANK_SCORE, start, end)
}

/**
 * 获取帐号附近的几个记录
 * @param {string} accountid
 * @return {Promise<unknown>}
 */
export async function getRankNear(accountid: string) {
  let rank: any = await getAccountRank(accountid)
  if (rank == null) {
    return []
  }
  let total: any = await new RedisClient().zcard(BaseConst.RANK_SCORE)
  let start = 0
  let count = 2
  let end = start + count
  if (rank > 0 && rank < total - 1) {
    start = rank - 1
    end = start + count
  } else if (rank == total - 1) {
    start = rank - 2
    end = total
  }
  start = start < 0? 0 : start
  end = end > total - 1 ? total - 1 : end
  let datas: any = await getRankList(start, end)
  let results: any[] = []
  for (let i = 0, l = datas.length; i < l; i += 2) {
    results.push({
      rank: start + (i / 2),
      accountid: datas[i],
      score: datas[i + 1] << 0
    })
  }
  return results
}



